#!/usr/bin/perl

use strict;
use warnings;

my $FORMAT_VERSION = '2012_01_04_01';
my @version_parts = split(/_/, $FORMAT_VERSION);

my $version_string = pack("nCCC", int($version_parts[0]),
		int($version_parts[1]),
		int($version_parts[2]),
		int($version_parts[3]));

my $version_int = $version_parts[0] . $version_parts[1] .
		$version_parts[2] . $version_parts[3];

my %POWERUP_TYPES = (
	'none' => 0,
	'life' => 1,
	'coin' => 2
);

my %tile_table = ();
my $k=0;
$tile_table{' '} = $k++;
foreach ('a'..'z'){
	$tile_table{$_} = $k++;
}
foreach ('A'..'Z'){
	$tile_table{$_} = $k++;
}
foreach (0..9){
	$tile_table{$_} = $k++;
}

undef $k;

my $gamefile;
my $name;

sub fail {
	print STDERR "\n " . "##########################################################";
	print STDERR "\n " . "#                                                        #";
	print STDERR "\n " . "# ERROR: Compiling failed, game file may now be corrupt! #";
	print STDERR "\n " . "#                                                        #";
	print STDERR "\n " . "##########################################################";
	print STDERR "\n\n";
	print STDERR " $_[0]\n\n" if(@_ && length($_[0]));
	exit 1;
}

sub put_padded_string{
	my ($text, $width) = @_;
	return $text . "\x00"x($width-length($text));
}

sub write_tile {
	my ($tiledir, $j, %tile) = @_;

	print $gamefile "t";
	print $gamefile put_padded_string($tile{name}, 16);
	print $gamefile ($tile{passable} eq 'true'?"\x01":"\x00");
	print $gamefile ($tile{deadly} eq 'true'?"\x01":"\x00");
	
	open(TILEPNM, "tilesets/$tiledir/tile_" . ($j<10?"0$j":"$j") . ".pnm") || die "Could not open tile PNM: $!";
	binmode(TILEPNM);
	seek(TILEPNM, 0, 2);
	my $tilesize = tell(TILEPNM);
	if($tilesize > 0xFFFF){
		fail "Tile image too large";
	}
	print $gamefile pack("n", $tilesize);
	seek(TILEPNM, 0, 0);
	while(<TILEPNM>){
		print $gamefile $_;
	}
	close TILEPNM;
	
	print "\n";
}


sub collect_tilesets {
	opendir(TILEROOT, "tilesets") || die "Could not open tileset root: $!";
	my @tiledirs = readdir TILEROOT; #XXX: Assumes sorted
	closedir TILEROOT;
	foreach my $tiledir (@tiledirs){
		next if($tiledir=~/^\./ || !-d "tilesets/$tiledir");
		
		my %tile = ();
		
		print $gamefile 'T';
		open(TILESET, '<', "tilesets/$tiledir/tileset.c2t") || die "Could not open tileset file: $!";
		
		my $j=0;
		my $intile = 0;
		foreach my $line (<TILESET>){
			chomp $line;
			$line =~ s/^\s*//;
			$line =~ s/\s*$//;
			next if(!length($line) || $line=~/^#/);
			if(!$intile){
				if($line=~/^name\s*=\s*([a-zA-Z0-9_\- ]{1,16})$/){
					print "Tileset name: $1\n";
					print $gamefile put_padded_string($1, 16);
				}elsif($line=~/^\[([0-9]{2})\]$/){
					$intile = 1;
					%tile = ();
				}else{
					fail "Syntax error in tileset '$tiledir' ($j): '$line'";
				}
			}else{
				if($line=~/^\[([0-9]{2})\]$/ && $1 == $j+1){

					(exists($tile{name}) && exists($tile{passable}) && exists($tile{deadly})) ||
						fail "Incomplete tile definition in tileset '$tiledir' ($j): '$line'";

					write_tile($tiledir, $j, %tile);
					$j++;
					
					%tile = ();
				}elsif($line=~/^name\s*=\s*([a-zA-Z0-9_\- ]{1,16})$/){
					print "Tile name: $1\n";
					$tile{name} = $1;
				}elsif($line=~/^passable\s*=\s*(true|false)$/){
					print "Passable = $1\n";
					$tile{passable} = $1;
				}elsif($line=~/^deadly\s*=\s*(true|false)$/){
					print "Deadly = $1\n";
					$tile{deadly} = $1;
				}else{
					fail "Syntax error in tileset '$tiledir' ($j): '$line'";
				}
			}
		}
		
		(exists($tile{name}) && exists($tile{passable}) && exists($tile{deadly})) ||
			fail "Incomplete tile definition in tileset '$tiledir' ($j)";
		
		write_tile($tiledir, $j, %tile);
		
		print $gamefile "\xFF";
		print "\n";
		fail "Tileset '$tiledir' is empty" if(!$intile || $j == 0);
		close TILESET;
	}
}

sub collect_entities {
	opendir(ENTITYROOT, "entities") || die "Could not open entity root: $!";
	my @entitydirs = readdir ENTITYROOT;
	closedir ENTITYROOT;
	my $j=0;
	my %entity = ();
	foreach my $entitydir (@entitydirs){
		next if($entitydir=~/^\./ || !-d "entities/$entitydir");
		%entity = ();
		print $gamefile 'E';
		open(ENTITY, '<', "entities/$entitydir/entity.c2e") || die "Could not open entity file: $!";
		
		foreach my $line (<ENTITY>){
			chomp $line;
			$line =~ s/^\s*//;
			next if(!length($line) || $line=~/^#/);
			
			if($line=~/^name\s*=\s*([a-zA-Z0-9_\- ]{1,16})$/){
				$entity{name} = $1;
				print "Entity name: $1\n";
			}elsif($line=~/^width\s*=\s*([0-9]{1,3})$/){
				$entity{width} = $1;
				print "Width: $1\n";
			}elsif($line=~/^height\s*=\s*([0-9]{1,3})$/){
				$entity{height} = $1;
				print "Height: $1\n";
			}elsif($line=~/^hitbox_left\s*=\s*([0-9]{1,3})$/){
				$entity{hitbox_left} = $1;
				print "Hitbox offset left: $1\n";
			}elsif($line=~/^hitbox_top\s*=\s*([0-9]{1,3})$/){
				$entity{hitbox_top} = $1;
				print "Hitbox offset top: $1\n";
			}elsif($line=~/^hitbox_right\s*=\s*([0-9]{1,3})$/){
				$entity{hitbox_right} = $1;
				print "Hitbox offset right: $1\n";
			}elsif($line=~/^hitbox_bottom\s*=\s*([0-9]{1,3})$/){
				$entity{hitbox_bottom} = $1;
				print "Hitbox offset bottom: $1\n";
			}elsif($line=~/^collectable\s*=\s*(true|false)$/){
				$entity{collectable} = $1;
				print "Collectable = $1\n";
			}elsif($line=~/^powerup_type\s*=\s*([a-zA-Z0-9_\- ]+)$/ && exists($POWERUP_TYPES{$1})){
				$entity{powerup_type} = $1;
				print "Powerup type: $1\n";
			}elsif($line=~/^hostile\s*=\s*(true|false)$/){
				$entity{hostile} = $1;
				print "Hostile = $1\n";
			}elsif($line=~/^has_ai\s*=\s*(true|false)$/){
				$entity{has_ai} = $1;
				print "Has AI = $1\n";
			}elsif($line=~/^can_move\s*=\s*(true|false)$/){
				$entity{can_move} = $1;
				print "Can move = $1\n";
			}elsif($line=~/^follows_player\s*=\s*(true|false)$/){
				$entity{follows_player} = $1;
				print "Follows player = $1\n";
			}elsif($line=~/^avoids_cliffs\s*=\s*(true|false)$/){
				$entity{avoids_cliffs} = $1;
				print "Avoids cliffs = $1\n";
			}else{
				fail "Syntax error in entity '$entitydir': '$line'";
			}
		}
		
		(exists($entity{name}) && exists($entity{width}) && exists($entity{height}) &&
				exists($entity{hitbox_left}) && exists($entity{hitbox_top}) &&
				exists($entity{hitbox_right}) && exists($entity{hitbox_bottom}) &&
				exists($entity{powerup_type}) && exists($entity{collectable}) &&
				exists($entity{hostile}) && exists($entity{has_ai}) &&
				exists($entity{can_move}) && exists($entity{follows_player}) &&
				exists($entity{avoids_cliffs})) ||
			fail "Incomplete entity definition in entity '$entitydir'";
		
		print $gamefile put_padded_string($entity{name}, 16);
		print $gamefile pack("C", $entity{width});
		print $gamefile pack("C", $entity{height});
		print $gamefile pack("C", $entity{hitbox_left});
		print $gamefile pack("C", $entity{hitbox_top});
		print $gamefile pack("C", $entity{hitbox_right});
		print $gamefile pack("C", $entity{hitbox_bottom});
		print $gamefile pack("C", $POWERUP_TYPES{$entity{powerup_type}});
		
		my $flags = (
				($entity{collectable}    eq 'true'?(1<<5):0) |
				($entity{hostile}        eq 'true'?(1<<4):0) |
				($entity{has_ai}         eq 'true'?(1<<3):0) |
				($entity{can_move}       eq 'true'?(1<<2):0) |
				($entity{follows_player} eq 'true'?(1<<1):0) |
				($entity{avoids_cliffs}  eq 'true'?(1<<0):0)
		);
		
		print $gamefile pack("n", $flags);
		
		open(ENTITYPNM, "entities/$entitydir/entity.pnm") || die "Could not open entity PNM: $!";
		binmode(ENTITYPNM);
		seek(ENTITYPNM, 0, 2);
		my $entitysize = tell(ENTITYPNM);
		if($entitysize > 0xFFFF){
			fail "Entity image too large";
		}
		print $gamefile pack("n", $entitysize);
		seek(ENTITYPNM, 0, 0);
		while(<ENTITYPNM>){
			print $gamefile $_;
		}
		close ENTITYPNM;
		$j++;
		
		print "\n";
		
		print $gamefile "\xFF";
		#fail "Error in entity '$entitydir' (state:$state)" if($state != -1);
		close ENTITY;
	}
}

sub collect_levels {
	opendir(LEVELROOT, "levels") || die "Could not open level root: $!";
	my @leveldirs = readdir LEVELROOT;
	closedir LEVELROOT;
	foreach my $levelfile (@leveldirs){
		next if($levelfile=~/^\./ || !-f "levels/$levelfile" || $levelfile!~/.c2l$/);
		
		my %level = ();
		
		print $gamefile 'L';
		open(LEVEL, '<', "levels/$levelfile") || die "Could not open level file: $!";
		
		my $level_length = 0;
		my $j=0;
		foreach my $line (<LEVEL>){
			chomp $line;
			$line =~ s/^\s*//;
			next if(!length($line) || $line=~/^#/);

			if($line=~/^name\s*=\s*([a-zA-Z0-9_\- ]{1,16})$/){
				print "Level name: $1\n";
				$level{name} = $1;
			}elsif($line=~/^start_x\s*=\s*([0-9]{1,2})$/){
				print "Start x: $1\n";
				$level{start_x} = $1;
			}elsif($line=~/^start_y\s*=\s*([0-9]{1,2})$/){
				print "Start y: $1\n";
				$level{start_y} = $1;
			}elsif($line=~/^tileset_id\s*=\s*([0-9]{1,2})$/){
				print "Tileset ID: $1\n";
				$level{tileset_id} = $1;
			}elsif($line=~/^:([a-zA-Z0-9 ]+)$/){
				my $row = $1;
				if($j == 0){
					$level{width} = length($row);
					print "Width: $level{width}\n";
				}else{
					if(length($row) != $level{width}){
						fail "Level length not uniform (row " .($j+1). " = ".length($row)." tiles, while level length is set to $level{width})";
					}
				}
				$level{rows}[$j] = $row;
				$j++;
			}else{
				fail "Syntax error in level '$levelfile': '$line'";
			}
		}
		
		(exists($level{name}) && exists($level{start_x}) && exists($level{start_y}) &&
				exists($level{tileset_id}) && exists($level{width}) && exists($level{rows})) ||
			fail "Incomplete level definition in level '$levelfile'";
		
		print $gamefile put_padded_string($level{name}, 16);
		print $gamefile pack("C", $level{start_x});
		print $gamefile pack("C", $level{start_y});
		print $gamefile pack("C", $level{tileset_id});
		
		print $gamefile pack("n", $level{width});
		
		foreach my $row (@{$level{rows}}){
			my @chars = split(//, $row);
			foreach(@chars){
				$_ = $tile_table{$_};
				if($_ >= 52){
					$_ -= 52;
					$_ |= 0b01000000;
				}elsif($_ >= 26){
					$_ -= 26;
					$_ |= 0b10000000;
				}
				$_ = pack("C", $_);
				print $gamefile $_;
			}
		}
		
		print "\n";
		
		print $gamefile "\xFF";
		#fail "Error in level '$levelfile' (state:$state)" if($state != -1);
		close LEVEL;
	}
}

if(@ARGV != 1){
	print "Usage: compile_gamefile.pl <gamefile>\n";
	exit;
}

print <<EOF;
compile_gamefile.pl
Copyright (c) 2011-2012, Chris Johan Smeele (cjsx9g)
All rights reserved.

EOF

my $gamedescription_filename = $ARGV[0];

if(!-f $gamedescription_filename || !-r $gamedescription_filename){
	fail "Game description file '$gamedescription_filename' does not exist or isn't readable.\n";
	exit 1;
}

print "Output file format version: $FORMAT_VERSION ($version_int)\n";

my $lives;

my $gamedescription_file;
open($gamedescription_file, '<', $gamedescription_filename) || fail "Could not open game description file for reading: $!";
my $i=0;
my $got_version = 0;
while(<$gamedescription_file>){
	$i++;
	chomp;
	s/^\s*//;
	next if(!length($_) || /^#/);
	if(!$got_version){
	if(/^format_version:([0-9]{4})_([0-9]{2})_([0-9]{2})_([0-9]{2})$/){
		#my $game_format_version = pack("nCCC",
		#		int($1), int($2), int($3), int($4));
		my $game_format_int = $1.$2.$3.$4;
		print "Input file format version:  $1_$2_$3_$4 ($game_format_int)\n";
		if((int $game_format_int) < (int $version_int)){
			fail "Game format version is too old, please update your files or use an older version of compile_gamefile.pl";
		}elsif((int $game_format_int) > (int $version_int)){
			fail "Game format version is too new, please update compile_gamefile.pl or downgrade your game files";
		}
		$got_version = 1;
	}else{
		fail "Expected format version as first variable at $gamedescription_filename, line $i.";
	}
	}else{
		if(/^name\s*=\s*([a-zA-Z0-9_\- ]{1,16})$/){
			print "Game name: '$1'\n";
			$name = $1;
		}elsif(/^lives\s*=\s*([0-9]{1,2})$/){
			print "Lives: $1\n\n";
			$lives = $1;
		}else{
			fail "Syntax error at $gamedescription_filename, line $i.";
		}
	}
}
close $gamedescription_file;

if(!defined $name || !length($name) || !defined $lives || !length($lives)){
	fail "Incomplete game definition in game description file $gamedescription_filename";
}

my $gamefilename = "$name";
$gamefilename =~ s/ /_/g;
$gamefilename .= '.c2d';

open($gamefile, '>', $gamefilename) || fail "Could not open game file for writing: $!";
binmode($gamefile);
print $gamefile "C2D\x05\x39" . $version_string . "\x00";
print $gamefile put_padded_string($name, 16);
print $gamefile pack("C", $lives);

print "--------\nTilesets\n\n";
collect_tilesets;
print "--------\nEntities\n\n";
collect_entities;
print "--------\nLevels\n\n";
collect_levels;

print $gamefile "\xff";


close $gamefile;








